from collections import defaultdict
from uuid import uuid4


class SessionExchange:
    def __init__(self):
        self._instances = defaultdict(list)

    def last(self, instance_type_name):
        return self._instances.get(instance_type_name, None).pop()

    def new(self, instance_type_name):
        oid = uuid4()
        self._instances[instance_type_name].append(oid)
        return oid

    def clear_all(self):
        for k in self._instances.keys():
            self._instances[k] = []



from .hub import Hub
from .helpers import expose, subscribe_commands, auto_subscribe  # noqa

__version__ = "0.1.3-alpha"

pub = Hub()


def subscribe(topic):
    def _subscribe(func):
        pub.subscribe(func, topic)
        return func
    return _subscribe



import weakref
from inspect import getfullargspec, ismethod, isfunction


class PubSubError(Exception):
    pass


def callable_class(obj):
    # all classes have an implicity __call__ function!!!!
    # fer fuck's sake
    call_func = getattr(obj, "__call__", None)
    # we need to call bound on the obj to eliminate functions
    # we also need to eliminate functions as they can
    # have __call__ attributes
    return (
        call_func is not None
        and not isinstance(obj, type)
        and not isfunction(obj)
        and not bound(obj)
        and bound(call_func)
    )


def bound(obj):
    _self = getattr(obj, "__self__", None)
    return _self is not None


def has_self_arg(obj):
    # classes with __init__ methods have self
    if isinstance(obj, type):
        # this is a class - which is what we don't want
        return False
    # instances of classes will cause this to error
    try:
        return "self" in getfullargspec(obj).args
    except TypeError:  # noqa
        return False


def requires_weakref(obj):
    return (bound(obj) and has_self_arg(obj)) or callable_class(obj)


def acceptable(obj):
    # must be callable
    if not callable(obj):
        return False
    # must not be a class
    if isinstance(obj, type):
        return False

    if ismethod(obj):
        return True

    if callable_class(obj):
        return True

    if not bound(obj) and has_self_arg(obj):
        return False

    return True


def apply_ref(obj):

    if ismethod(obj):
        return weakref.WeakMethod(obj)

    return weakref.ref(obj)



import logging
from .event import PubSubError, Event
from .util import SessionExchange

log = logging.getLogger("beercraft")
log_msg = logging.getLogger("beercraft.messages")


class HubSingleton(type):
    _hub = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._hub:
            cls._hub[cls] = super(HubSingleton, cls).__call__(*args, **kwargs)

        return cls._hub[cls]


class Hub(metaclass=HubSingleton):
    def __init__(self):
        self._events = {}
        self._missing_topics = {}
        self.oid = SessionExchange()

    def add_topic(self, topic):
        """ A topic can only be added once and only
        has the one Event object mapped to it.
        """
        if not isinstance(topic, str):
            msg = "Topic argument must be a string! (Given: {0})"
            raise PubSubError(msg.format(topic))

        if not self.has(topic):
            self._events[topic] = Event(topic)

    def subscribe(self, subscriber, topic):

        if not callable(subscriber):
            msg = 'Subscriber ({0}) to topic "{1}" must be a callable.'
            raise PubSubError(msg.format(subscriber, topic))

        self.add_topic(topic)
        log.debug("subscribe {0} to {1}".format(subscriber, topic))
        self._events[topic] += subscriber

    def send_message(self, topic, **kwargs):
        if self.has(topic):
            try:
                log_msg.debug("calling {0} with {1}".format(topic, kwargs.keys()))
                event = self._events[topic]
                log.debug(event)
                event(**kwargs)
            except Exception as ex:  # pylint: disable=broad-except
                if self.has(f'{topic}.exceptions'):
                    self.send_message(f'{topic}.exceptions', error=ex)

                elif self.has('exceptions'):
                    self.send_message('exceptions', error=ex)

                else:
                    raise ex
        else:
            log.debug("{0} has no subscribers.".format(topic))
            self._missing_topics.setdefault(topic, 0)
            self._missing_topics[topic] += 1

    def mk_msg_func(self, topic, **kwargs):
        def _call():
            self.send_message(topic, **kwargs)

        return _call

    def has(self, topic):
        return topic in self._events

    def clear_all(self):
        self._events.clear()
        self._missing_topics.clear()
        self.oid.clear_all()

    def topic_list(self):
        return list(self._events.keys())

    def missing_topic_list(self):
        return list(self._missing_topics.keys())



import logging
from .types import PubSubError, acceptable, apply_ref


log = logging.getLogger("beercraft.event")


class EventError(PubSubError):
    def __init__(self, handler, ex):
        super(EventError, self).__init__(ex)
        self.handler = handler
        self.ex = ex

    def __str__(self):
        return "EventError: {0}".format(self.ex)


class EventErrorHandler:
    def __init__(self):
        self.halt_current = True
        self.halt_parent = True
        self._handler = self._raise

    @property
    def handler(self):
        return self._handler

    @handler.setter
    def handler(self, value):
        self._handler = value

    def _raise(self, exception):
        raise EventError(self, exception)

    def handle_error(self, exception):
        self._handler(exception)


class Event:
    def __init__(self, eid):
        self.eid = eid
        self._notify_list = []
        self._clean = False
        self._errors = EventErrorHandler()

    @property
    def halt_on_error(self):
        return self._errors.halt_current

    @halt_on_error.setter
    def halt_on_error(self, value):
        self._errors.halt_current = value

    def set_error_handler(self, handler):
        self._errors.handler = handler

    def _iter_ref(self):
        """Return the WeakMethod object.
        """
        for listener_ref in self._notify_list:
            yield listener_ref

    def _iter_callables(self):
        """Return the call to WeakMethod object.
        Gives the referenced object.
        """
        for listener_ref in self._notify_list:
            yield listener_ref()

    def register(self, listener):
        """ register function!
        """
        if not acceptable(listener):
            msg = "{} is not a valid beercraft pub subscriber!"
            raise PubSubError(msg.format(str(listener)))
        lref = apply_ref(listener)
        self._notify_list.append(lref)

    def unregister(self, listener):
        """ unregister fuction!
        """
        references = self._iter_ref()
        for listener_ref in references:
            if listener_ref() == listener:
                self._notify_list.remove(listener_ref)
                references.close()

    def notify(self, *args, **kwargs):
        """ Call all listeners.
        """
        callables = self._iter_callables()
        for listener_obj in callables:
            log.debug("calling {0}".format(listener_obj))
            if listener_obj:
                try:
                    listener_obj(*args, **kwargs)
                except Exception as ex:  # pylint: disable=broad-except
                    try:
                        self._errors.handle_error(ex)
                    finally:
                        if self._errors.halt_current:
                            callables.close()
                            raise EventError(self._errors, ex)
            else:
                self._clean = True

        self._cleanup()

    def _cleanup(self):
        """remove any null references.
        """
        self._notify_list = [c for c in self._notify_list if c() is not None]

    def clear(self):
        """remove all listners.
        """
        self._notify_list = []

    def __contains__(self, other):
        callables = self._iter_callables()
        contains = False
        for obj in callables:
            if obj == other:
                callables.close()
                contains = True
        return contains

    def __call__(self, *args, **kwargs):
        self.notify(*args, **kwargs)

    def __iadd__(self, listener):
        self.register(listener)
        return self

    def __isub__(self, listener):
        self.unregister(listener)
        return self



import re
import logging
import warnings
from inspect import getmembers
from functools import wraps
from .types import PubSubError
from .hub import Hub


log = logging.getLogger("beercraft")


INTERP_RX = re.compile(r"\{\d?\}")


def topic_formatter(topic_string):

    requires_sub = len(INTERP_RX.findall(topic_string)) > 0

    def _topic_formatter(substitution=None):
        if requires_sub:
            if not substitution:
                err_msg = "Topic `{0}` must be supplied a frame id!"
                raise PubSubError(err_msg.format(topic_string))
            return topic_string.format(substitution)
        return topic_string

    return _topic_formatter


def expose(topic, logger=None):

    if not isinstance(topic, str):
        raise PubSubError("A beercraft topic must be string!")

    topic = topic_formatter(topic)

    if logger:
        _log = logger
    else:
        _log = log

    if not hasattr(_log, "debug"):
        raise PubSubError("logger must have a `debug` method.")

    def expose_decorator(func):
        """Wraps the base method
        """

        @wraps(func)
        def wrapper_(*args, **kwargs):
            _log.debug(str(func))
            return func(*args, **kwargs)

        wrapper_.topic = topic
        wrapper_.pubsub = True
        return wrapper_

    return expose_decorator


def subscribe_commands(obj):
    """Scan objects for methods that have been decorated with the
    ExposeCommand and subscribe them to pubsub.

    """
    pubsubhub = Hub()
    pub_id = getattr(obj, "__pub_id__", "")

    if getattr(obj, "__auto_subscribed__", False) is True:
        klass = type(obj).__name__
        msg = (
            "Class {} is decorated with the beercraft.auto_subcribe "
            "function. Calling `subscribe_commands` is redundant."
        )
        warnings.warn(msg.format(klass), category=RuntimeWarning)
        return

    def is_exposed(obj):
        return hasattr(obj, "pubsub") and getattr(obj, "pubsub") is True

    exposed_mems = [name for name, inst in getmembers(obj) if is_exposed(inst)]
    exposed_mems.sort()
    log.debug("beercraft.subscrib_commands({})".format(obj))
    log.debug("   object id: {}".format(pub_id))
    for mem_name in exposed_mems:
        log.debug(f"   subscribing member `{mem_name}` of {mem_name}")
        inst = getattr(obj, mem_name)
        topic = inst.topic(pub_id)
        pubsubhub.subscribe(inst, topic)


def auto_subscribe(type_name=None):
    def _auto_subscribe(klass):
        """Class decorator that automates the subscription
        of new instances of a class.
        """
        if not isinstance(klass, type):
            raise PubSubError(
                "The `auto_subscribe` class decorator "
                "can only be applied to a type. "
                "Received `{}`".format(klass)
            )
        pub = Hub()
        org_init = klass.__init__

        def __init__(self, *args, **kwargs):

            if not hasattr(self, "__pub_id__") and type_name:
                self.__pub_id__ = pub.oid.last(type_name)

            org_init(self, *args, **kwargs)

            subscribe_commands(self)

            self.__auto_subscribed__ = True

        klass.__auto_subscribed__ = False

        klass.__init__ = __init__
        return klass

    return _auto_subscribe
